#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Module de génération de fichiers CSV à partir de la base de données
PostgreSQL du CERDD

Metadata dans pyproject.toml et dans README.md
"""

try:
    from pocli_csv_export._version import __version__
except ImportError:
    __version__ = "1.0.1"

import csv
import os
import sys
from contextlib import ExitStack

from psycopg2 import connect
from slugify import slugify

# REGLAGES ET PARAMETRAGES
from pocli_csv_export.settings_defaults import (
    CSV_MAIN,
    CSV_PARAM,
    DESTINATION_FOLDER,
    DSN,
    EXPORT_COLUMN,
    FILIERES,
    TYPE_TER,
)

try:
    from pocli_csv_export.settings import *  # noqa
except ImportError:
    sys.stderr.write(
        "Veuillez configurer settings.py en "
        "vous inspirant de settings_defaults.py\n"
    )
    exit()

FIELDS_NAME, COLUMNS_TITLE = list(zip(*EXPORT_COLUMN))

FILIERES_LIST = [
    (filiere, energy_type, indicators)
    for filiere, val0 in FILIERES.items()
    for energy_type, indicators in val0.items()
]

MTX_IND_TYPE = {
    indicator: ind_type
    for filiere, val0 in FILIERES.items()
    for energy_type, indicators in val0.items()
    for indicator, ind_type in indicators.items()
}


class CsvWriters:
    def __init__(self, files_dict):
        self.csvwriters = {
            fname: csv.writer(files_dict[fname], **CSV_PARAM)
            for fname in files_dict
        }

    def write_fil(self, filiere, row):
        self.csvwriters[CSV_MAIN].writerow(row)
        self.csvwriters[filiere].writerow(row)

    def write_all(self, row):
        for csv_writer in self.csvwriters.values():
            csv_writer.writerow(row)


class Query:
    def __init__(self, cur, query, const_dict=None, func_dict=None):
        self.cur = cur
        self.query = query
        self.const_dict = const_dict
        self.func_dict = func_dict
        self.colname = None
        self.rows = None

    def execute(self):
        self.cur.execute(self.query)
        self.colname = {e.name: idx for idx, e in enumerate(cur.description)}
        self.rows = self.cur.fetchall()
        return self

    def get_rows(self):
        if self.colname is None:
            self.execute()
        for row in self.rows:
            yield Row(self.colname, row, self.const_dict, self.func_dict)


class Row:
    def __init__(self, colname, row, const_dict=None, func_list=None):
        self.colname = colname
        const_dict = const_dict or {}
        row_dict = dict(zip(colname, map(self._to_python, row)))
        self.row = {**const_dict, **row_dict}
        if func_list is None:
            return
        for attr, func, args_name in func_list:
            args = [self.row.get(arg_name) for arg_name in args_name]
            self.row[attr] = func(attr, *args)

    def _to_python(self, val):
        return '' if val is None else str(val).replace('.', ',')

    def __getattr__(self, attr):
        if attr not in self.colname:
            raise AttributeError(f"{attr} not in {self}")
        return self._to_python(self.row[attr])


def get_main_query(indicators):
    return f"""
    SELECT
     v.annee,
     t.id_ter,
     t.type_ter,
     t.lib_ter,
     va.dep,
     v.valeur,
     t.geoloc_long,
     t.geoloc_lat,
     v.id_indic
    FROM
     indic.valeur v
      INNER JOIN indic.description d ON v.id_indic = d.id_indic::text
      LEFT OUTER JOIN trans.v_territoires t ON v.id_ter = t.id_ter
      LEFT OUTER JOIN trans.v_appartenance va ON v.id_ter = va.id_ter
    WHERE
     d.niveau_indic = 3
     AND t.type_ter in {tuple(TYPE_TER)}
     AND v.id_indic in {tuple(indicators)}
    ORDER BY
     v.annee,
     t.type_ter,
     t.id_ter
    """


def accumulate(row_acc, row):
    row_acc = row_acc or {}
    for key, val in row.row.items():
        if row_acc.get(key, '') == '':
            row_acc[key] = val
    return row_acc


def same_row_id(row_tail, row_head):
    NAT_PKEY = ('annee', 'id_ter', 'type_ter', 'filiere', 'energy_type')
    pkey_row_tail = [row_tail.row[key] for key in NAT_PKEY]
    pkey_row_head = [row_head.row[key] for key in NAT_PKEY]
    return pkey_row_tail == pkey_row_head


def flat_row(row_acc):
    return [row_acc.get(fname) for fname in FIELDS_NAME]


class FuncHelper:
    def __init__(self, cur):
        self.cur = cur
        self.dep_dict = self.get_dep_dict()
        self.func_list = [
            ('nom_dep', self.get_nom_dep, ['dep']),
            ('ind_type', self.get_ind_type, ['id_indic']),
            ('nb_install', self.val_or_not, ['ind_type', 'valeur']),
            ('production', self.val_or_not, ['ind_type', 'valeur']),
            ('puissance', self.val_or_not, ['ind_type', 'valeur']),
        ]

    def get_dep_dict(self):
        query = """
        SELECT id_ter, lib_ter
        FROM trans.v_territoires
        WHERE type_ter = 'DEPARTEMENT'
        """
        return {
            row.id_ter: row.lib_ter
            for row in Query(self.cur, query).get_rows()
        }

    def get_nom_dep(self, attr, dep):
        return self.dep_dict.get(dep, '')

    def get_ind_type(self, attr, id_indic):
        return MTX_IND_TYPE[id_indic]

    def val_or_not(self, attr, ind_type, valeur):
        return valeur if attr == ind_type else ''


def main(cur, csvwriters):
    func_list = FuncHelper(cur).func_list
    csvwriters.write_all(COLUMNS_TITLE)
    for filiere, energy_type, indicators in FILIERES_LIST:
        var_dict = {'filiere': filiere, 'energy_type': energy_type}
        str_query = get_main_query(indicators)
        main_query = Query(cur, str_query, var_dict, func_list).execute()
        print(f"{filiere}/{energy_type}: {len(main_query.rows)}")
        iter_rows = main_query.get_rows()
        try:
            row_tail = next(iter_rows)
        except StopIteration:
            pass  # cas liste vide
        else:
            row_acc = None
            for row_head in iter_rows:
                row_acc = accumulate(row_acc, row_tail)
                if not same_row_id(row_tail, row_head):
                    csvwriters.write_fil(filiere, flat_row(row_acc))
                    row_acc = None
                row_tail = row_head
            # cas de la derniere ecriture, cas liste 1 ligne
            row_acc = accumulate(row_acc, row_tail)
            csvwriters.write_fil(filiere, flat_row(row_acc))


if __name__ == '__main__':

    def get_fd(fname):
        fname = fname if fname == CSV_MAIN else slugify(fname)
        path = os.path.join(DESTINATION_FOLDER, f"{fname}.csv")
        return open(path, 'w', encoding='utf-8-sig')

    with ExitStack() as stack:
        conn = stack.enter_context(connect(**DSN))
        cur = stack.enter_context(conn.cursor())
        files_dict = {
            fname: stack.enter_context(get_fd(fname))
            for fname in [CSV_MAIN] + list(FILIERES)
        }
        csvwriters = CsvWriters(files_dict)
        main(cur, csvwriters)
    # end with
